package server;

import java.net.*;
import java.io.*;

/**
 * The socket worker keeps a (TCP) connection to the client open as
 * long as there are no I/O and no protocol errors. This is a
 * "stateless" worker, since there is only one instance (but several
 * threads with a reference to it). The information is entirely in the
 * arguments.
 */
class SocketWorker implements Task, ProtocolConstants {

	/** The server that we're managing. */
	private UserIDServer server;

	/** How long are we waiting for an answer. */
	private int maxBlockTime;

	/* some error messages */
	public static final String ioError = "IOError in communcation, closing.";
	public static final String protocolError = "Format error from client, closing.";
	public static final String closingError = "Error while closing socket.";
	public static final String idCollision = "Different user tried to register with same ID, denying.";
	public static final String starting = "Starting a new connection.";

	SocketWorker(UserIDServer server) {
		this.server = server;
		maxBlockTime = server.LIFETIME * 1000 + 500;
	}

	/**
	 * A call to this method represents a connection with a client.
	 * It keeps waiting for a REGISTER and sends back a USERLIST.
	 */
	public void doTask(Object[] args) {
		/* set up the socket options */
		Socket socket = (Socket)args[0];
		try {
			socket.setSoTimeout(maxBlockTime);
		} catch (SocketException se) {
			System.err.println(se.getMessage());
			finish(socket);
			return;
		}

		/* create the I/O streams */
		DataInputStream in;
		DataOutputStream out;
		try {
			in = new DataInputStream(socket.getInputStream());
			out = new DataOutputStream(socket.getOutputStream());
		} catch (IOException ioe) {
			ioe.printStackTrace();
			finish(socket);
			return;
		}

		if (UserIDServer.DEBUG)
			System.err.println(starting);

		/* communicate */
		InetAddress address = socket.getInetAddress();

		try {
			byte type = in.readByte();
			switch (type) {
				case REGISTER :
					{
						registerUser(in, address);
						sendUserList(out);
						break;
					}
				case GETUSERS :
					{
						sendUserList(out);
					}
			}
		} catch (IOException ioe) {
			if (UserIDServer.DEBUG) {
				ioe.printStackTrace();
			}
		} catch (Exception e) {
			// any other errors (NullPointer, ArrayIndexOutOfBounds...)
			if (UserIDServer.DEBUG) {
				e.printStackTrace();
			}
		}

		/* clean up */
		// Don't need to close the socket since the client does it...
		try {
			in.close();
			out.close();
			socket.close();
		} catch (IOException ioe) {
			System.err.println(closingError);
		}
	}

	/**
	 * Parses the user data and registers the user with the server.
	 * Format for the registration: REGISTER(1) port(2) name length(1)
	 * name (< 256)
	 *
	 * @param stream the InputStream of the socket connected with the
	 * user
	 */
	private void registerUser(DataInputStream stream, InetAddress address) throws IOException {
		User user = null;

		/* parse data */
		int port = read16(stream);
		int namelen = stream.readByte() & 0xff;
		byte[] buffer = new byte[namelen];

		for (int i = 0; i < namelen; i++) {
			buffer[i] = (byte)stream.read();
		}

		//		stream.read(buffer); // Doesn't work nice! If a user sends his name byte by byte, stream.read(buffer) returns after the first byte was received! It only works, if the user sends it all at once!
		String name = new String(buffer);
		byte[] addr = address.getAddress();
		if ((addr.length < 4) && UserIDServer.DEBUG) {
			System.out.println("Ooops, bad address: " + address);
			return;
		}

		/* register user */
		user = server.getUser(addr, port);
		if (user == null) {
			if (UserIDServer.DEBUG)
				System.out.println("creating new user...");
			user = new User(name, addr, port);
			try {
				UserIDServer.fileWriter.write("+++ ");
				UserIDServer.fileWriter.write((new java.util.Date(System.currentTimeMillis())).toString());
				UserIDServer.fileWriter.write(" *** ");
				UserIDServer.fileWriter.write(user.toString() + "\n");
				UserIDServer.fileWriter.flush();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		} else {
			user.setName(name); // in case it changed
			try {
				UserIDServer.fileWriter.write("*** ");
				UserIDServer.fileWriter.write((new java.util.Date(System.currentTimeMillis())).toString());
				UserIDServer.fileWriter.write(" *** ");
				UserIDServer.fileWriter.write(user.toString() + "\n");
				UserIDServer.fileWriter.flush();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		server.registerUser(user);
	}

	/**
	 * Sends the user back a list with all currently registered users.
	 * Format: USERLIST(1) number of users(2) + each user, where each
	 * user is: namelen(1) name(var) ip-address(4) port(2).
	 *
	 * @param stream the OutputStream of the socket connected with the
	 * user
	 */
	private void sendUserList(DataOutputStream stream) throws IOException {
		User[] users = server.getUsers();
		int number = users.length;

		stream.writeByte(USERLIST);
		write16(number, stream);

		for (int i = 0; i < users.length; i++) {
			User u = users[i];
			String name = u.getName();
			stream.writeByte(name.length() & 0xff);
			stream.write(name.getBytes(), 0, name.length());
			byte[] buffer = u.getAddress();
			stream.write(buffer, 0, buffer.length);
			write16(u.getPort(), stream);
		}
	}

	/** Writes a 2-byte number to the stream, higher order first. */
	private void write16(int value, DataOutputStream stream) throws IOException {
		stream.writeByte((byte) ((value & 0xff00) >> 8));
		stream.writeByte((byte)value & 0xff);
	}

	/** Reads a 2-byte number from the stream, higher order first. */
	private int read16(DataInputStream stream) throws IOException {
		return (int) ((stream.readByte() & 0xff) << 8) | (stream.readByte() & 0xff);
	}

	/** Compares two byte-arrays. */
	private boolean addressesEqual(byte[] a1, byte[] a2) {
		if (a1.length != a2.length)
			return false;

		for (int i = 0; i < a1.length; i++) {
			if (a1[i] != a2[i])
				return false;
		}
		return true;
	}

	private void finish(Socket socket) {
		try {
			socket.close();
		} catch (IOException ioe) {
			System.err.println(closingError);
		}
	}
}
